@@ -1,10 +1,12 @@
+from typing import Optional
+
 import numpy as np
-from scipy.special import gammaln
+from scipy.special import gammaln  # type: ignore
 
 import pydeseq2.utils
 
 
-def vec_nb_nll(counts, mu, alpha):
+def vec_nb_nll(counts: np.ndarray, mu: np.ndarray, alpha: float) -> np.ndarray:
     """Return the negative log-likelihood of a negative binomial.
 
     Vectorized version.
@@ -14,7 +16,7 @@ def vec_nb_nll(counts, mu, alpha):
     counts : ndarray
         Observations.
 
-    mu : float
+    mu : ndarray
         Mean of the distribution.
 
     alpha : float
@@ -23,7 +25,7 @@ def vec_nb_nll(counts, mu, alpha):
 
     Returns
     -------
-    float
+    ndarray
         Negative log likelihood of the observations counts following
         :math:`NB(\\mu, \\alpha)`.
     """
@@ -51,17 +53,17 @@ def vec_nb_nll(counts, mu, alpha):
 
 
 def grid_fit_alpha(
-    counts,
-    design_matrix,
-    mu,
-    alpha_hat,
-    min_disp,
-    max_disp,
-    prior_disp_var=None,
-    cr_reg=True,
-    prior_reg=False,
-    grid_length=100,
-):
+    counts: np.ndarray,
+    design_matrix: np.ndarray,
+    mu: np.ndarray,
+    alpha_hat: float,
+    min_disp: float,
+    max_disp: float,
+    prior_disp_var: Optional[float] = None,
+    cr_reg: bool = True,
+    prior_reg: bool = False,
+    grid_length: int = 100,
+) -> float:
     """Find best dispersion parameter.
 
     Performs 1D grid search to maximize negative binomial log-likelihood.
@@ -90,13 +92,13 @@ def grid_fit_alpha(
         Prior dispersion variance.
 
     cr_reg : bool
-        Whether to use Cox-Reid regularization. (default: True).
+        Whether to use Cox-Reid regularization. (default: ``True``).
 
     prior_reg : bool
-        Whether to use prior log-residual regularization. (default: False).
+        Whether to use prior log-residual regularization. (default: ``False``).
 
     grid_length : int
-        Number of grid points. (default: 100).
+        Number of grid points. (default: ``100``).
 
     Returns
     -------
@@ -108,7 +110,7 @@ def grid_fit_alpha(
     max_log_alpha = np.log(max_disp)
     grid = np.linspace(min_log_alpha, max_log_alpha, grid_length)
 
-    def loss(log_alpha):
+    def loss(log_alpha: np.ndarray) -> np.ndarray:
         # closure to minimize
         alpha = np.exp(log_alpha)
         W = mu[:, None] / (1 + mu[:, None] * alpha)
@@ -121,6 +123,11 @@ def grid_fit_alpha(
                 )[1]
             )
         if prior_reg:
+            if prior_disp_var is None:
+                raise NotImplementedError(
+                    "If you use prior_reg, you need to specify prior_disp_var"
+                )
+
             reg += (np.log(alpha) - np.log(alpha_hat)) ** 2 / (2 * prior_disp_var)
         return vec_nb_nll(counts, mu, alpha) + reg
 
@@ -138,15 +145,15 @@ def grid_fit_alpha(
 
 
 def grid_fit_beta(
-    counts,
-    size_factors,
-    design_matrix,
-    disp,
-    min_mu=0.5,
-    grid_length=60,
-    min_beta=-30,
-    max_beta=30,
-):
+    counts: np.ndarray,
+    size_factors: np.ndarray,
+    design_matrix: np.ndarray,
+    disp: float,
+    min_mu: float = 0.5,
+    grid_length: int = 60,
+    min_beta: float = -30,
+    max_beta: float = 30,
+) -> np.ndarray:
     """Find best LFC parameter.
 
     Perform 2D grid search to maximize negative binomial
@@ -157,7 +164,7 @@ def grid_fit_beta(
     counts : ndarray
         Raw counts for a given gene.
 
-    size_factors : pandas.Series
+    size_factors : ndarray
         DESeq2 normalization factors.
 
     design_matrix : ndarray
@@ -170,24 +177,24 @@ def grid_fit_beta(
         Lower threshold for dispersion parameters.
 
     grid_length : int
-        Number of grid points. (default: 100).
+        Number of grid points. (default: ``100``).
 
-    min_beta : int
-        Lower-bound on LFC. (default: 30).
+    min_beta : float
+        Lower-bound on LFC. (default: ``30``).
 
-    max_beta : int
-        Upper-bound on LFC. (default: 30).
+    max_beta : float
+        Upper-bound on LFC. (default: ``30``).
 
     Returns
     -------
-    float
+    ndarray
         Fitted LFC parameter.
     """
     x_grid = np.linspace(min_beta, max_beta, grid_length)
     y_grid = np.linspace(min_beta, max_beta, grid_length)
     ll_grid = np.zeros((grid_length, grid_length))
 
-    def loss(beta):
+    def loss(beta: np.ndarray) -> np.ndarray:
         # closure to minimize
         mu = np.maximum(size_factors[:, None] * np.exp(design_matrix @ beta.T), min_mu)
         return vec_nb_nll(counts, mu, disp) + 0.5 * (1e-6 * beta**2).sum(1)
@@ -217,17 +224,17 @@ def grid_fit_beta(
 
 
 def grid_fit_shrink_beta(
-    counts,
-    offset,
-    design_matrix,
-    size,
-    prior_no_shrink_scale,
-    prior_scale,
-    scale_cnst,
-    grid_length=60,
-    min_beta=-30,
-    max_beta=30,
-):
+    counts: np.ndarray,
+    offset: np.ndarray,
+    design_matrix: np.ndarray,
+    size: np.ndarray,
+    prior_no_shrink_scale: float,
+    prior_scale: float,
+    scale_cnst: float,
+    grid_length: int = 60,
+    min_beta: float = -30,
+    max_beta: float = 30,
+) -> np.ndarray:
     """Find best LFC parameter.
 
     Performs 2D grid search to maximize MAP negative binomial
@@ -257,24 +264,24 @@ def grid_fit_shrink_beta(
         Scaling factor for the optimization.
 
     grid_length : int
-        Number of grid points. (default: 100).
+        Number of grid points. (default: ``100``).
 
     min_beta : int
-        Lower-bound on LFC. (default: 30).
+        Lower-bound on LFC. (default: ``30``).
 
     max_beta : int
-        Upper-bound on LFC. (default: 30).
+        Upper-bound on LFC. (default: ``30``).
 
     Returns
     -------
-    float
+    ndarray
         Fitted MAP LFC parameter.
     """
     x_grid = np.linspace(min_beta, max_beta, grid_length)
     y_grid = np.linspace(min_beta, max_beta, grid_length)
     ll_grid = np.zeros((grid_length, grid_length))
 
-    def loss(beta):
+    def loss(beta: np.ndarray) -> float:
         # closure to minimize
         return (
             pydeseq2.utils.nbinomFn(
