@@ -1,15 +1,21 @@
 import multiprocessing
 from math import floor
 from pathlib import Path
+from typing import List
+from typing import Literal
+from typing import Optional
+from typing import Tuple
+from typing import Union
+from typing import cast
 
 import numpy as np
 import pandas as pd
-from scipy.linalg import solve
-from scipy.optimize import minimize
-from scipy.special import gammaln
-from scipy.special import polygamma
-from scipy.stats import norm
-from sklearn.linear_model import LinearRegression
+from scipy.linalg import solve  # type: ignore
+from scipy.optimize import minimize  # type: ignore
+from scipy.special import gammaln  # type: ignore
+from scipy.special import polygamma  # type: ignore
+from scipy.stats import norm  # type: ignore
+from sklearn.linear_model import LinearRegression  # type: ignore
 
 import pydeseq2
 from pydeseq2.grid_search import grid_fit_alpha
@@ -18,15 +24,16 @@ from pydeseq2.grid_search import grid_fit_shrink_beta
 
 
 def load_example_data(
-    modality="raw_counts",
-    dataset="synthetic",
-    debug=False,
-    debug_seed=42,
-):
-    """Load synthetic or TCGA data (gene raw counts or clinical) for a given cancer type.
+    modality: Literal["raw_counts", "clinical"] = "raw_counts",
+    dataset: Literal["synthetic"] = "synthetic",
+    debug: bool = False,
+    debug_seed: int = 42,
+) -> pd.DataFrame:
+    """Load synthetic example data.
 
-    May load either clinical or rna-seq data.The synthetic data is part of this
-    repo, but TCGA data should be downloaded as per the instructions in `datasets/`.
+    May load either clinical or rna-seq data. For now, this function may only return the
+    synthetic data provided as part of this repo, but new datasets might be added in the
+    future.
 
     Parameters
     ----------
@@ -34,15 +41,16 @@ def load_example_data(
         Data modality. "raw_counts" or "clinical".
 
     dataset : str
-        The cancer type for which to return gene expression data.
+        The dataset for which to return gene expression data.
         If "synthetic", will return the synthetic data that is used for CI unit tests.
-        Otherwise, must be a valid TCGA dataset. (default: "synthetic").
+        (default: ``"synthetic"``).
 
     debug : bool
-        If true, subsample 10 samples and 100 genes at random. (default: False).
+        If true, subsample 10 samples and 100 genes at random.
+        (Note that the "synthetic" dataset is already 10 x 100.) (default: ``False``).
 
     debug_seed : int
-        Seed for the debug mode. (default: 42).
+        Seed for the debug mode. (default: ``42``).
 
     Returns
     -------
@@ -55,67 +63,52 @@ def load_example_data(
     )
 
     assert dataset in [
-        "synthetic",
-        "TCGA-BRCA",
-        "TCGA-COAD",
-        "TCGA-LUAD",
-        "TCGA-LUSC",
-        "TCGA-PAAD",
-        "TCGA-PRAD",
-        "TCGA-READ",
-        "TCGA-SKCM",
-    ], (
-        "The dataset argument must be one of the following: "
-        "synthetic, TCGA-BRCA, TCGA-COAD, TCGA-LUAD, TCGA-LUSC, "
-        "TCGA-PAAD, TCGA-PRAD, TCGA-READ, TCGA-SKCM"
-    )
+        "synthetic"
+    ], "The dataset argument must be one of the following: synthetic."
 
     # Load data
     datasets_path = Path(pydeseq2.__file__).parent.parent / "datasets"
 
     if dataset == "synthetic":
         path_to_data = datasets_path / "synthetic"
-
-        if modality == "raw_counts":
-            df = pd.read_csv(
-                path_to_data / "test_counts.csv",
-                sep=",",
-                index_col=0,
-            ).T
-        elif modality == "clinical":
-            df = pd.read_csv(
-                path_to_data / "test_clinical.csv",
-                sep=",",
-                index_col=0,
+        if Path(path_to_data).is_dir():
+            # Cast the Paths to strings to have coherent types wrt to the url case (that
+            # does not handle Paths), else mypy throws an error.
+            path_to_data_counts = str(path_to_data / "test_counts.csv")
+            path_to_data_clinical = str(path_to_data / "test_clinical.csv")
+        else:
+            # if the path does not exist (as is the case in RDT) load it from github
+            url_to_data = (
+                "https://raw.githubusercontent.com/owkin/"
+                "PyDESeq2/main/datasets/synthetic/"
             )
+            path_to_data_counts = url_to_data + "/test_counts.csv"
+            path_to_data_clinical = url_to_data + "/test_clinical.csv"
 
-    else:
-        path_to_data = datasets_path / "tcga_data"
         if modality == "raw_counts":
             df = pd.read_csv(
-                path_to_data / "Gene_expressions" / f"{dataset}_raw_RNAseq.tsv.gz",
-                compression="gzip",
-                sep="\t",
+                path_to_data_counts,
+                sep=",",
                 index_col=0,
             ).T
         elif modality == "clinical":
             df = pd.read_csv(
-                path_to_data / "Clinical" / f"{dataset}_clinical.tsv.gz",
-                compression="gzip",
-                sep="\t",
+                path_to_data_clinical,
+                sep=",",
                 index_col=0,
             )
 
     if debug:
+        # TODO: until we provide a larger dataset, this option is useless
         # subsample 10 samples and 100 genes
         df = df.sample(n=10, axis=0, random_state=debug_seed)
         if modality == "raw_counts":
-            df = df.sample(n=100, axis=1, random_state=debug_seed)
+            df = df.sample(n=100, axis="index", random_state=debug_seed)
 
     return df
 
 
-def test_valid_counts(counts_df):
+def test_valid_counts(counts_df: pd.DataFrame) -> None:
     """Test that the count matrix contains valid inputs.
 
     More precisely, test that inputs are non-negative integers.
@@ -138,8 +131,12 @@ def test_valid_counts(counts_df):
 
 
 def build_design_matrix(
-    clinical_df, design_factors="condition", ref=None, expanded=False, intercept=True
-):
+    clinical_df: pd.DataFrame,
+    design_factors: Union[str, List[str]] = "condition",
+    ref: Optional[str] = None,
+    expanded: bool = False,
+    intercept: bool = True,
+) -> pd.DataFrame:
     """Build design_matrix matrix for DEA.
 
     Only single factor, 2-level designs are currently supported.
@@ -152,21 +149,21 @@ def build_design_matrix(
         DataFrame containing clinical information.
         Must be indexed by sample barcodes.
 
-    design_factors : str or list[str]
+    design_factors : str or list
         Name of the columns of clinical_df to be used as design_matrix variables.
-        (default: "condition").
+        (default: ``"condition"``).
 
     ref : str
         The factor to use as a reference. Must be one of the values taken by the design.
         If None, the reference will be chosen alphabetically (last in order).
-        (default: None).
+        (default: ``None``).
 
     expanded : bool
         If true, use one column per category. Else, use a single column.
-        (default: False).
+        (default: ``False``).
 
     intercept : bool
-        If true, add an intercept (a column containing only ones). (default: True).
+        If true, add an intercept (a column containing only ones). (default: ``True``).
 
     Returns
     -------
@@ -208,14 +205,17 @@ def build_design_matrix(
             columns=[col for col in design_matrix.columns[::-2]], axis=1, inplace=True
         )  # Drops every other column, starting from the last
     if intercept:
-        design_matrix.insert(0, "intercept", 1.0)
+        design_matrix.insert(0, "intercept", 1)
     return design_matrix
 
 
-def dispersion_trend(normed_mean, coeffs):
+def dispersion_trend(
+    normed_mean: Union[float, np.ndarray],
+    coeffs: Union["pd.Series[float]", np.ndarray],
+) -> Union[float, np.ndarray]:
     r"""Return dispersion trend from normalized counts.
 
-     :math:`a_1/ \mu + a_0`.
+    :math:`a_1/ \mu + a_0`.
 
     Parameters
     ----------
@@ -227,37 +227,67 @@ def dispersion_trend(normed_mean, coeffs):
 
     Returns
     -------
-    float
+    float or ndarray
         Dispersion trend :math:`a_1/ \mu + a_0`.
     """
-    if type(coeffs) == pd.Series:
+    if isinstance(coeffs, pd.Series):
         return coeffs["a0"] + coeffs["a1"] / normed_mean
     else:
         return coeffs[0] + coeffs[1] / normed_mean
 
 
-def nb_nll(counts, mu, alpha):
-    """Negative log-likelihood of a negative binomial.
+def nb_nll(counts: np.ndarray, mu: np.ndarray, alpha: float) -> float:
+    """Negative log-likelihood of a negative binomial of parameters ``mu`` and ``alpha``.
 
     Unvectorized version.
 
+    Mathematically, if ``counts`` is a vector of counting entries :math:`y_i`
+    then the likelihood of each entry :math:`y_i` to be drawn from a negative
+    binomial :math:`NB(\\mu, \\alpha)` is [1]
+
+    .. math::
+        p(y_i | \\mu, \\alpha) = \\frac{\\Gamma(y_i + \\alpha^{-1})}{
+            \\Gamma(y_i + 1)\\Gamma(\\alpha^{-1})
+        }
+        \\left(\\frac{1}{1 + \\alpha \\mu} \\right)^{1/\\alpha}
+        \\left(\\frac{\\mu}{\\alpha^{-1} + \\mu} \\right)^{y_i}
+
+    As a consequence, assuming there are :math:`n` entries,
+    the total negative log-likelihood for ``counts`` is
+
+    .. math::
+        \\ell(\\mu, \\alpha) = \\frac{n}{\\alpha} \\log(\\alpha) +
+            \\sum_i \\left \\lbrace
+            - \\log \\left( \\frac{\\Gamma(y_i + \\alpha^{-1})}{
+            \\Gamma(y_i + 1)\\Gamma(\\alpha^{-1})
+        } \\right)
+        + (\\alpha^{-1} + y_i) \\log (\\alpha^{-1} + y_i)
+        - y_i \\log \\mu
+            \\right \\rbrace
+
+    This is implemented in this function.
+
     Parameters
     ----------
     counts : ndarray
         Observations.
 
-    mu : float
-        Mean of the distribution.
+    mu : ndarray
+        Mean of the distribution :math:`\\mu`.
 
     alpha : float
-        Dispersion of the distribution,
-        s.t. the variance is :math:`\\mu + \\alpha * \\mu^2`.
+        Dispersion of the distribution :math:`\\alpha`,
+        s.t. the variance is :math:`\\mu + \\alpha \\mu^2`.
 
     Returns
     -------
     float
         Negative log likelihood of the observations counts
         following :math:`NB(\\mu, \\alpha)`.
+
+    Notes
+    -----
+    [1] https://en.wikipedia.org/wiki/Negative_binomial_distribution
     """
 
     n = len(counts)
@@ -273,7 +303,7 @@ def nb_nll(counts, mu, alpha):
     )
 
 
-def dnb_nll(counts, mu, alpha):
+def dnb_nll(counts: np.ndarray, mu: np.ndarray, alpha: float) -> float:
     """Gradient of the negative log-likelihood of a negative binomial.
 
     Unvectorized.
@@ -311,57 +341,59 @@ def dnb_nll(counts, mu, alpha):
 
 
 def irls_solver(
-    counts,
-    size_factors,
-    design_matrix,
-    disp,
-    min_mu=0.5,
-    beta_tol=1e-8,
-    min_beta=-30,
-    max_beta=30,
-    optimizer="L-BFGS-B",
-    maxiter=250,
-):
+    counts: np.ndarray,
+    size_factors: np.ndarray,
+    design_matrix: np.ndarray,
+    disp: float,
+    min_mu: float = 0.5,
+    beta_tol: float = 1e-8,
+    min_beta: float = -30,
+    max_beta: float = 30,
+    optimizer: Literal["BFGS", "L-BFGS-B"] = "L-BFGS-B",
+    maxiter: int = 250,
+) -> Tuple[np.ndarray, np.ndarray, np.ndarray, bool]:
     r"""Fit a NB GLM wit log-link to predict counts from the design matrix.
 
     See equations (1-2) in the DESeq2 paper.
 
     Parameters
     ----------
-    counts : pandas.Series
+    counts : ndarray
         Raw counts for a given gene.
 
-    size_factors : pandas.Series
+    size_factors : ndarray
         Sample-wise scaling factors (obtained from median-of-ratios).
 
-    design_matrix : pandas.DataFrame
+    design_matrix : ndarray
         Design matrix.
 
-    disp : pandas.Series
-        Gene-wise dispersion priors.
+    disp : float
+        Gene-wise dispersion prior.
 
     min_mu : float
-        Lower bound on estimated means, to ensure numerical stability. (default: 0.5).
+        Lower bound on estimated means, to ensure numerical stability.
+        (default: ``0.5``).
 
     beta_tol : float
         Stopping criterion for IRWLS:
-        :math:`abs(dev - old_dev) / (abs(dev) + 0.1) < beta_tol`. (default: 1e-8).
+        :math:`\vert dev - dev_{old}\vert / \vert dev + 0.1 \vert < \beta_{tol}`.
+        (default: ``1e-8``).
 
-    min_beta : int
-        Lower-bound on LFC. (default: -30).
+    min_beta : float
+        Lower-bound on LFC. (default: ``-30``).
 
-    max_beta : int
-        Upper-bound on LFC. (default: -30).
+    max_beta : float
+        Upper-bound on LFC. (default: ``-30``).
 
     optimizer : str
         Optimizing method to use in case IRLS starts diverging.
         Accepted values: 'BFGS' or 'L-BFGS-B'.
         NB: only 'L-BFGS-B' ensures that LFCS will
-        lay in the [min_beta, max_beta] range. (default: 'L-BFGS-B').
+        lay in the [min_beta, max_beta] range. (default: ``'L-BFGS-B'``).
 
     maxiter : int
         Maximum number of IRLS iterations to perform before switching to L-BFGS-B.
-        (default: 250).
+        (default: ``250``).
 
     Returns
     -------
@@ -369,26 +401,20 @@ def irls_solver(
         Fitted (basemean, lfc) coefficients of negative binomial GLM.
 
     mu: ndarray
-        Means estimated from size factors and beta:
-        :math:`\\mu = s_{ij} \\exp(\\beta^t design_matrix)`.
+        Means estimated from size factors and beta: :math:`\mu = s_{ij} \exp(\beta^t X)`.
 
     H: ndarray
         Diagonal of the :math:`W^{1/2} X (X^t W X)^-1 X^t W^{1/2}` covariance matrix.
+
+    converged: bool
+        Whether IRLS or the optimizer converged. If not and if dimension allows it,
+        perform grid search.
     """
 
     assert optimizer in ["BFGS", "L-BFGS-B"]
 
     num_vars = design_matrix.shape[1]
-
-    # converting to numpy for speed
-    if type(counts) == pd.Series:
-        counts = counts.values
-    if type(size_factors) == pd.Series:
-        size_factors = size_factors.values
-    if type(design_matrix) == pd.Series:
-        X = design_matrix.values
-    else:
-        X = design_matrix
+    X = design_matrix
 
     # if full rank, estimate initial betas for IRLS below
     if np.linalg.matrix_rank(X) == num_vars:
@@ -397,7 +423,7 @@ def irls_solver(
         beta_init = solve(R, Q.T @ y)
         beta = beta_init
 
-    dev = 1000
+    dev = 1000.0
     dev_ratio = 1.0
 
     ridge_factor = np.diag(np.repeat(1e-6, num_vars))
@@ -406,7 +432,6 @@ def irls_solver(
     converged = True
     i = 0
     while dev_ratio > beta_tol:
-
         W = mu / (1.0 + mu * disp)
         z = np.log(mu / size_factors) + (counts - mu) / mu
         H = (X.T * W) @ X + ridge_factor
@@ -415,12 +440,12 @@ def irls_solver(
 
         if sum(np.abs(beta_hat) > max_beta) > 0 or i >= maxiter:
             # If IRLS starts diverging, use L-BFGS-B
-            def f(beta):
+            def f(beta: np.ndarray) -> float:
                 # closure to minimize
                 mu_ = np.maximum(size_factors * np.exp(X @ beta), min_mu)
                 return nb_nll(counts, mu_, disp) + 0.5 * (ridge_factor @ beta**2).sum()
 
-            def df(beta):
+            def df(beta: np.ndarray) -> np.ndarray:
                 mu_ = np.maximum(size_factors * np.exp(X @ beta), min_mu)
                 return (
                     -X.T @ counts
@@ -472,17 +497,17 @@ def irls_solver(
 
 
 def fit_alpha_mle(
-    counts,
-    design_matrix,
-    mu,
-    alpha_hat,
-    min_disp,
-    max_disp,
-    prior_disp_var=None,
-    cr_reg=True,
-    prior_reg=False,
-    optimizer="L-BFGS-B",
-):
+    counts: np.ndarray,
+    design_matrix: np.ndarray,
+    mu: np.ndarray,
+    alpha_hat: float,
+    min_disp: float,
+    max_disp: float,
+    prior_disp_var: Optional[float] = None,
+    cr_reg: bool = True,
+    prior_reg: bool = False,
+    optimizer: Literal["BFGS", "L-BFGS-B"] = "L-BFGS-B",
+) -> Tuple[float, bool]:
     """Estimate the dispersion parameter of a negative binomial GLM.
 
     Note: it is possible to pass counts, design_matrix and mu arguments in the form of
@@ -512,14 +537,14 @@ def fit_alpha_mle(
         Prior dispersion variance.
 
     cr_reg : bool
-        Whether to use Cox-Reid regularization. (default: True).
+        Whether to use Cox-Reid regularization. (default: ``True``).
 
     prior_reg : bool
-        Whether to use prior log-residual regularization. (default: False).
+        Whether to use prior log-residual regularization. (default: ``False``).
 
     optimizer : str
         Optimizing method to use. Accepted values: 'BFGS' or 'L-BFGS-B'.
-        (default: 'BFGS').
+        (default: ``'L-BFGS-B'``).
 
     Returns
     -------
@@ -533,11 +558,12 @@ def fit_alpha_mle(
     assert optimizer in ["BFGS", "L-BFGS-B"]
 
     if prior_reg:
+        # Note: assertion is not working when using numpy
         assert (
             prior_disp_var is not None
         ), "Sigma_prior is required for prior regularization"
 
-    def loss(log_alpha):
+    def loss(log_alpha: float) -> float:
         # closure to be minimized
         alpha = np.exp(log_alpha)
         W = mu / (1 + mu * alpha)
@@ -545,10 +571,13 @@ def fit_alpha_mle(
         if cr_reg:
             reg += 0.5 * np.linalg.slogdet((design_matrix.T * W) @ design_matrix)[1]
         if prior_reg:
+            if prior_disp_var is None:
+                raise ValueError("Sigma_prior is required for prior regularization")
+
             reg += (np.log(alpha) - np.log(alpha_hat)) ** 2 / (2 * prior_disp_var)
         return nb_nll(counts, mu, alpha) + reg
 
-    def dloss(log_alpha):
+    def dloss(log_alpha: float) -> float:
         # gradient closure
         alpha = np.exp(log_alpha)
         W = mu / (1 + mu * alpha)
@@ -563,13 +592,16 @@ def fit_alpha_mle(
                 ).sum()
             )
         if prior_reg:
+            if prior_disp_var is None:
+                raise ValueError("Sigma_prior is required for prior regularization")
+
             reg_grad += (np.log(alpha) - np.log(alpha_hat)) / (alpha * prior_disp_var)
         return dnb_nll(counts, mu, alpha) + reg_grad
 
     res = minimize(
         lambda x: loss(x[0]),
-        x0=[np.log(alpha_hat)],
-        jac=dloss,
+        x0=np.log(alpha_hat),
+        jac=lambda x: dloss(x[0]),
         method=optimizer,
         bounds=[(np.log(min_disp), np.log(max_disp))]
         if optimizer == "L-BFGS-B"
@@ -587,7 +619,7 @@ def fit_alpha_mle(
         )
 
 
-def trimmed_mean(x, trim=0.1, **kwargs):
+def trimmed_mean(x, trim: float = 0.1, **kwargs) -> Union[float, np.ndarray]:
     """Return trimmed mean.
 
     Compute the mean after trimming data of its smallest and largest quantiles.
@@ -598,7 +630,7 @@ def trimmed_mean(x, trim=0.1, **kwargs):
         Data whose mean to compute.
 
     trim : float
-        Fraction of data to trim at each end. (default: 0.1).
+        Fraction of data to trim at each end. (default: ``0.1``).
 
     **kwargs
         Keyword arguments, useful to pass axis.
@@ -623,7 +655,7 @@ def trimmed_mean(x, trim=0.1, **kwargs):
         return s[ntrim : n - ntrim].mean()
 
 
-def trimmed_cell_variance(counts, cells):
+def trimmed_cell_variance(counts: pd.DataFrame, cells: pd.Series) -> pd.Series:
     """Return trimmed variance of counts according to condition.
 
     Compute the variance after trimming data of its smallest and largest elements,
@@ -635,7 +667,7 @@ def trimmed_cell_variance(counts, cells):
     counts : pandas.DataFrame
         Sample-wise gene counts.
 
-    cells : pandas.DataFrame
+    cells : pandas.Series
         Cohort affiliation of each sample.
 
     Returns
@@ -648,7 +680,7 @@ def trimmed_cell_variance(counts, cells):
     trimratio = (1 / 3, 1 / 4, 1 / 8)
     # returns an index for the vector above for three sample size bins
 
-    def trimfn(x):
+    def trimfn(x: float) -> int:
         return 2 if x >= 23.5 else 1 if x >= 3.5 else 0
 
     ns = cells.value_counts()
@@ -671,10 +703,12 @@ def trimmed_cell_variance(counts, cells):
     return varEst.max(axis=1)
 
 
-def trimmed_variance(x, trim=0.125, axis=0):
+def trimmed_variance(
+    x: np.ndarray, trim: float = 0.125, axis: int = 0
+) -> Union[float, np.ndarray]:
     """Return trimmed variance.
 
-     Compute the variance after trimming data of its smallest and largest quantiles.
+    Compute the variance after trimming data of its smallest and largest quantiles.
 
     Parameters
     ----------
@@ -682,14 +716,14 @@ def trimmed_variance(x, trim=0.125, axis=0):
         Data whose trimmed variance to compute.
 
     trim : float
-        Fraction of data to trim at each end. (default: 0.125).
+        Fraction of data to trim at each end. (default: ``0.125``).
 
     axis : int
-        Dimension along which to compute variance. (default: 0).
+        Dimension along which to compute variance. (default: ``0``).
 
     Returns
     -------
-    ndarray
+    float or ndarray
         Trimmed variances.
     """
 
@@ -699,7 +733,12 @@ def trimmed_variance(x, trim=0.125, axis=0):
     return 1.51 * trimmed_mean(sqerror, trim=trim, axis=axis)
 
 
-def fit_lin_mu(counts, size_factors, design_matrix, min_mu=0.5):
+def fit_lin_mu(
+    counts: np.ndarray,
+    size_factors: np.ndarray,
+    design_matrix: np.ndarray,
+    min_mu: float = 0.5,
+) -> np.ndarray:
     """Estimate mean of negative binomial model using a linear regression.
 
     Used to initialize genewise dispersion models.
@@ -716,11 +755,11 @@ def fit_lin_mu(counts, size_factors, design_matrix, min_mu=0.5):
         Design matrix.
 
     min_mu : float
-        Lower threshold for fitted means, for numerical stability. (default: 0.5).
+        Lower threshold for fitted means, for numerical stability. (default: ``0.5``).
 
     Returns
     -------
-    float
+    ndarray
         Estimated mean.
     """
 
@@ -731,7 +770,14 @@ def fit_lin_mu(counts, size_factors, design_matrix, min_mu=0.5):
     return np.maximum(mu_hat, min_mu)
 
 
-def wald_test(design_matrix, disp, lfc, mu, ridge_factor, idx=-1):
+def wald_test(
+    design_matrix: np.ndarray,
+    disp: float,
+    lfc: np.ndarray,
+    mu: float,
+    ridge_factor: np.ndarray,
+    idx: int = -1,
+) -> Tuple[float, float, float]:
     """Run Wald test for differential expression.
 
     Computes Wald statistics, standard error and p-values from
@@ -745,7 +791,7 @@ def wald_test(design_matrix, disp, lfc, mu, ridge_factor, idx=-1):
     disp : float
         Dispersion estimate.
 
-    lfc : float
+    lfc : ndarray
         Log-fold change estimate (in natural log scale).
 
     mu : float
@@ -755,7 +801,7 @@ def wald_test(design_matrix, disp, lfc, mu, ridge_factor, idx=-1):
         Regularization factors.
 
     idx : int
-        Index of design factor (in design matrix). (default: -1).
+        Index of design factor (in design matrix). (default: ``-1``).
 
     Returns
     -------
@@ -775,23 +821,26 @@ def wald_test(design_matrix, disp, lfc, mu, ridge_factor, idx=-1):
     H = np.linalg.inv(M + ridge_factor)
     S = H @ M @ H
     # Evaluate standard error and Wald statistic
-    wald_se = np.sqrt(S[idx, idx])
-    wald_statistic = lfc[idx] / wald_se
+    wald_se: float = np.sqrt(S[idx, idx])
+    wald_statistic: float = lfc[idx] / wald_se
     wald_p_value = 2 * norm.sf(np.abs(wald_statistic))
     return wald_p_value, wald_statistic, wald_se
 
 
-def fit_rough_dispersions(counts, size_factors, design_matrix):
+def fit_rough_dispersions(
+    counts: np.ndarray, size_factors: np.ndarray, design_matrix: pd.DataFrame
+) -> np.ndarray:
     """ "Rough dispersion" estimates from linear model, as per the R code.
 
-    Used as initial estimates in DeseqDataSet._fit_MoM_dispersions.
+    Used as initial estimates in :meth:`DeseqDataSet.fit_genewise_dispersions()
+    <pydeseq2.dds.DeseqDataSet.fit_genewise_dispersions>`.
 
     Parameters
     ----------
-    counts : pandas.DataFrame
-        Raw counts. One column per gene, rows are indexed by sample barcodes.
+    counts : ndarray
+        Raw counts. One column per gene, one row per sample.
 
-    size_factors : pandas.Series
+    size_factors : ndarray
         DESeq2 normalization factors.
 
     design_matrix : pandas.DataFrame
@@ -800,14 +849,17 @@ def fit_rough_dispersions(counts, size_factors, design_matrix):
 
     Returns
     -------
-    pandas.Series
+    ndarray
         Estimated dispersion parameter for each gene.
     """
 
     num_samples, num_vars = design_matrix.shape
-    normed_counts = counts.div(size_factors, 0)
+
+    normed_counts = counts / size_factors[:, None]
+
     # Exclude genes with all zeroes
-    normed_counts = normed_counts.loc[:, ~(normed_counts == 0).all()]
+    normed_counts = normed_counts[:, ~(normed_counts == 0).all(axis=0)]
+
     reg = LinearRegression(fit_intercept=False)
     reg.fit(design_matrix, normed_counts)
     y_hat = reg.predict(design_matrix)
@@ -818,37 +870,41 @@ def fit_rough_dispersions(counts, size_factors, design_matrix):
     return np.maximum(alpha_rde, 0)
 
 
-def fit_moments_dispersions(counts, size_factors):
+def fit_moments_dispersions(counts: np.ndarray, size_factors: np.ndarray) -> np.ndarray:
     """Dispersion estimates based on moments, as per the R code.
 
-    Used as initial estimates in DeseqDataSet._fit_MoM_dispersions.
+    Used as initial estimates in :meth:`DeseqDataSet.fit_genewise_dispersions()
+    <pydeseq2.dds.DeseqDataSet.fit_genewise_dispersions>`.
 
     Parameters
     ----------
-    counts : pandas.DataFrame
-        Raw counts. One column per gene, rows are indexed by sample barcodes.
+    counts : ndarray
+        Raw counts. One column per gene, one row per sample.
 
-    size_factors : pandas.Series
+    size_factors : ndarray
         DESeq2 normalization factors.
 
     Returns
     -------
-    pandas.Series
+    ndarray
         Estimated dispersion parameter for each gene.
     """
-    normed_counts = counts.div(size_factors, 0)
+
+    normed_counts = counts / size_factors[:, None]
     # Exclude genes with all zeroes
-    normed_counts = normed_counts.loc[:, ~(normed_counts == 0).all()]
+    normed_counts = normed_counts[:, ~(normed_counts == 0).all(axis=0)]
     # mean inverse size factor
     s_mean_inv = (1 / size_factors).mean()
     mu = normed_counts.mean(0)
     sigma = normed_counts.var(0, ddof=1)
     # ddof=1 is to use an unbiased estimator, as in R
     # NaN (variance = 0) are replaced with 0s
-    return ((sigma - s_mean_inv * mu) / mu**2).fillna(0)
+    return np.nan_to_num((sigma - s_mean_inv * mu) / mu**2)
 
 
-def robust_method_of_moments_disp(normed_counts, design_matrix):
+def robust_method_of_moments_disp(
+    normed_counts: pd.DataFrame, design_matrix: pd.DataFrame
+) -> pd.Series:
     """Perform dispersion estimation using a method of trimmed moments.
 
     Used for outlier detection based on Cook's distance.
@@ -875,26 +931,28 @@ def robust_method_of_moments_disp(normed_counts, design_matrix):
         v = trimmed_cell_variance(normed_counts, design_matrix.iloc[:, -1].astype(int))
         # last argument is non-intercept.
     else:
-        v = trimmed_variance(normed_counts)
+        v = pd.Series(
+            trimmed_variance(normed_counts.values), index=normed_counts.columns
+        )
     m = normed_counts.mean(0)
     alpha = (v - m) / m**2
     # cannot use the typical min_disp = 1e-8 here or else all counts in the same
     # group as the outlier count will get an extreme Cook's distance
     minDisp = 0.04
-    alpha = np.maximum(alpha, minDisp)
+    alpha = cast(pd.Series, np.maximum(alpha, minDisp))
     return alpha
 
 
-def get_num_processes(n_cpus=None):
+def get_num_processes(n_cpus: Optional[int] = None) -> int:
     """Return the number of processes to use for multiprocessing.
 
     Returns the maximum number of available cpus by default.
 
     Parameters
     ----------
-    n_cpus : int
-        Desired number of cpus. If None, will return the number of available cpus.
-        (default: None).
+    n_cpus : int or None
+        Desired number of cpus. If ``None``, will return the number of available cpus.
+        (default: ``None``).
 
     Returns
     -------
@@ -914,15 +972,15 @@ def get_num_processes(n_cpus=None):
 
 
 def nbinomGLM(
-    design_matrix,
-    counts,
-    size,
-    offset,
-    prior_no_shrink_scale,
-    prior_scale,
+    design_matrix: np.ndarray,
+    counts: np.ndarray,
+    size: np.ndarray,
+    offset: np.ndarray,
+    prior_no_shrink_scale: float,
+    prior_scale: float,
     optimizer="L-BFGS-B",
-    shrink_index=1,
-):
+    shrink_index: int = 1,
+) -> Tuple[np.ndarray, np.ndarray, bool]:
     """Fit a negative binomial MAP LFC using an apeGLM prior.
 
     Only the LFC is shrinked, and not the intercept.
@@ -949,10 +1007,10 @@ def nbinomGLM(
 
     optimizer : str
         Optimizing method to use in case IRLS starts diverging.
-        Accepted values: 'L-BFGS-B', 'BFGS' or 'Newton-CG'. (default: 'Newton-CG').
+        Accepted values: 'L-BFGS-B', 'BFGS' or 'Newton-CG'. (default: ``'Newton-CG'``).
 
     shrink_index : int
-        Index of the LFC coordinate to shrink. (default: 1).
+        Index of the LFC coordinate to shrink. (default: ``1``).
 
     Returns
     -------
@@ -987,7 +1045,7 @@ def nbinomGLM(
     )
     scale_cnst = np.maximum(scale_cnst, 1)
 
-    def f(beta, cnst=scale_cnst):
+    def f(beta: np.ndarray, cnst: float = scale_cnst) -> float:
         # Function to optimize
         return (
             nbinomFn(
@@ -1003,7 +1061,7 @@ def nbinomGLM(
             / cnst
         )
 
-    def df(beta, cnst=scale_cnst):
+    def df(beta: np.ndarray, cnst: float = scale_cnst) -> np.ndarray:
         # Gradient of the function to optimize
         xbeta = design_matrix @ beta
         d_neg_prior = (
@@ -1017,7 +1075,7 @@ def nbinomGLM(
 
         return (d_neg_prior - d_nll) / cnst
 
-    def ddf(beta, cnst=scale_cnst):
+    def ddf(beta: np.ndarray, cnst: float = scale_cnst) -> np.ndarray:
         # Hessian of the function to optimize
         # Note: will only work if there is a single shrink index
         xbeta = design_matrix @ beta
@@ -1068,15 +1126,15 @@ def nbinomGLM(
 
 
 def nbinomFn(
-    beta,
-    design_matrix,
-    counts,
-    size,
-    offset,
-    prior_no_shrink_scale,
-    prior_scale,
-    shrink_index=1,
-):
+    beta: np.ndarray,
+    design_matrix: np.ndarray,
+    counts: np.ndarray,
+    size: np.ndarray,
+    offset: np.ndarray,
+    prior_no_shrink_scale: float,
+    prior_scale: float,
+    shrink_index: int = 1,
+) -> float:
     """Return the NB negative likelihood with apeGLM prior.
 
     Use for LFC shrinkage.
@@ -1105,7 +1163,7 @@ def nbinomFn(
         Prior variance for the intercept.
 
     shrink_index : int
-        Index of the LFC coordinate to shrink. (default: 1).
+        Index of the LFC coordinate to shrink. (default: ``1``).
 
     Returns
     -------
